//|----------------------------------------------|
//| Slow/Haste/Safe/Shell Timers Hack            |
//| by: madsiur                                  |
//| version: 1.0a                                |
//| Released on: February 16th, 2022             |
//| apply to: FF3us 1.0 (no header)              |
//|----------------------------------------------|

// assemble with bass v14, e.g. "bass -o rom.smc status-timers-stop.asm"
// https://www.romhacking.net/utilities/794/

//|----------------------------------------------|
//| RAM changes (previously unused)              |
//|----------------------------------------------|

// $B8: current entity's expired status byte
// bit 0: Stop (original)
// bit 1: Reflect (original)
// bit 2: Freeze (original)
// bit 3: Sleep (original)
// bit 4: Haste (new)
// bit 5: Slow (new)
// bit 6: Shell (new)
// bit 7: Safe (new)

// $3E38,x: current entity's Slow timer
// $3E39,x: current entity's Haste timer
// $3E88,x: current entity's Shell timer
// $3E89,x: current entity's Safe timer

//|----------------------------------------------|
//| timers length (reference)                    |
//|----------------------------------------------|

// Sleep: $12
// Stop: $12
// Reflect: $1A
// Freeze: $22

// you can change those constants if needed

constant haste_timer($1E)
constant slow_timer($1E)
constant shell_timer($1E)
constant safe_timer($1E)

//|----------------------------------------------|
//| functions                                    |
//|----------------------------------------------|

arch snes.cpu

// set base and origin
macro seek(variable offset) {
    origin (offset & $3FFFFF)
    base offset
}

//|----------------------------------------------|
//| bank $C2 code (seibaby, assassin, madsiur)   |
//|----------------------------------------------|

// set status function pointers

seek($C246D4)
	dw set_slow             // set Slow pointer
	dw set_haste            // set Haste pointer
	
seek($C246DA)
	dw set_shell            // set Shell pointer
	dw set_safe            	// set Safe pointer


// Independent Stop code (seibaby)
// https://www.ff6hacking.com/forums/thread-3389.html

seek($C24680)
lda.b #$48                	// time until Stop wears off x 4

seek($C25A83)
	lda.w $3A91             // lower byte of battle time counter equivalent
	inc $3A91               // increment that counter equiv.
	and.b #$0F    
	cmp.b #$0A    
	bcs other               // branch if lower nibble of time counter was >= #$0A
							// otherwise, A now corresponds to one of the 10
							// onscreen entities.
	asl 
	tax 
	lda $3AA0,x
	lsr 
	bcc exit_b              // exit if entity not present in battle
	lda $3AF1,x             // get Stop timer, originally set to #$12
							// at C2/467D
	beq no_stop             // branch if it's 0, as that *should* mean the
							// entity is not stopped.

	dec $3AF1,x             // decrement Stop timer
	bne exit_b              // did it JUST run down on this tick?  if not, exit
	jmp expire_stop         // go set up Stop to wear off
	nop                     // note that I could have fit the entire patch in this
							// function with "STA $B8 / BRA $5B30", but the latter
							// instruction would've been so cheap and ugly that I'd
							// have had to kill myself.
no_stop:
	clc 
	lda $3ADC,x             // Timer that determines how often timers and time-based
							// events will countdown and happen for this entity.
	adc $3ADD,x             // Add ATB multiplier (set at C2/09D2: 64 normally,
							// 32 if Slowed, 84 if Hasted)
	sta $3ADC,x
	bcc exit_b              // exit if timer didn't meet or exceed 256

	lda $3AA0,x
	bit #$10                // is entity Wounded, Petrified, or Stopped, or is
							// somebody else under the influence of Quick?
	bne exit_b              // exit if any are true
	lda $3B05,x             // condemned counter - originally set at 09B4.
							// to be clear, this counter is "one off" from the
							// actual numerals you'll see onscreen:
							// 00 value = numerals disabled
							// 01 value = numerals at "00", 02 = "01", 03 = "02", etc
	cmp.b #$02
	bcc lbl_5AC9            // branch if counter < 2.  [i.e. numerals < "01",
							// meaning they're "00" or disabled.]
	dec 
	sta $3B05,x             // decrement counter
	dec                     // just think of this second "DEC" as an optimized
							// "CMP #$01", as we're not altering the counter.
	bne lbl_5AC9            // branch if counter now != 1  [i.e. numerals != "00"]

	jsr $5BC7               // cast Doom when countdown numerals reach 0
lbl_5AC9:
	jsr $5C1B               // check if character runs from combat
	jsr $5B4F               // trigger Poison, Seizure, Regen, Phantasm, or
							// tentacle Drain damage
	tdc 
	jsr check_timers        // decrement Reflect, Freeze, and Sleep timers if
							// applicable, and check if any have worn off
	inc $3AF0,x             // advance this entity's pointer to the next entry
							// in the 5AEA function table
	lda $3AF0,x
	txy                     // preserve X in Y
	and.b #$07              // convert it to 0-7, wrapping as necessary
	asl 
	tax 
	jmp ($5AEA,x)           // determine which periodic/damage healing type
							// will be checked on this entity's next tick.
other:
	sbc #$0A                // should only be reached if $3A91 AND 15 was >= 10
							// [i.e. not corresponding to any specific entity]
							// at start of function.  now subtract 10.
	asl 
	tax 
	jmp ($5AFA,x)
	tyx                     // restore X from Y
	exit_b:
	rts

// check timers and queue removal (madsiur)

seek($C25B06)
check_timers:
	sta.b $B8
    jsl check_timers_ex 	// call relocated and expanded function
    beq exit_a          	// branch if no status has expired
queue_removal:
    lda.b #$29          	// queue status removal function
    jmp $4E91
    exit_a:
    rts

// original Slow/Haste code at $C24669-$C24677

// set Slow and start timer

set_slow:
	jsl set_slow_ex
	bra brn_haste           // mark Haste to be cleared

// set Haste and start timer

set_haste:
	jsl set_haste_ex
	brn_haste:
	tsb.b $F6               // clear Slow if Haste set or clear Haste if slow set
	lda.w #$0004            // ATB gauge constant needs to be updated bit
	jmp $464C               // set bit to $3204,y

// original Shell/Safe code is simply a RTS

// set Shell and start timer

set_shell:
	php					
	jsl set_shell_ex
	plp
	rts

// set Safe and start timer

set_safe:
	php
	jsl set_safe_ex
	plp
	rts

// Premature Continuation fix (assassin)
// http://assassin17.brinkster.net/patches.htm#anchor27

// set to check for Poison on this entity's next tick

seek($C25B3B)
	lda.b #$10
	bra reuse               // finish this function's work by using its near-twin's
							// code.

// Added code to mark Stop for removal upon its timer's expiration

expire_stop:
	sta $B8                 // flag Stop for removal)
	bra queue_removal       // queue Stop to be removed. but leave Reflect, Freeze,
							// Haste, Slow, Safe, Shell and Sleep timers untouched.
							// let them resume on the next tick, just like the entity's 
							// Condemned countdown and periodic Regen/Seizure/etc.
							// do, which is the sensible behavior. 
							// after all, time is supposed to unfreeze *after* Stop wears off.
	nop
	rts                     // never reached, but just in case


// set to check for Regen, Seizure, Phantasm on this entity's next tick
// identical results to original function, but reordered slightly

	lda.b #$08
reuse:
	tyx
	ora $3E4C,x
	sta $3E4C,x             // Set bit 3 of $3E4C,X .  or bit 4 if we
							// reached here from Function C2/5B3B.
	rts

// remove status when timers have expired (madsiur)

seek($C2518A)
	jsl remove_status_ex	// call relocated and expanded function
	lda.b #$04              // original code
	tsb.w $11A4             // set lift status bit
	jmp $3167

//|----------------------------------------------|
//| outside bank $C2 code (madsiur)              |
//|----------------------------------------------|

seek($EEAF01)			// you can change this free space offset
						// as long as it's not in bank $C2
						// and that you have 202 ($CA) bytes of free space

// set Slow and start timer

scope set_slow_ex: {
	sep #$20                // 8-bit A
	lda.b #slow_timer       // Slow timer value
	sta $3E38,y             // set as timer
	rep #$20                // 16-bit A
	lda.w #$0008            // Haste bit
	rtl
}

// set Haste and start timer 

scope set_haste_ex: {
	sep #$20                // 8-bit A
	lda.b #haste_timer      // Haste timer value
	sta $3E39,y             // set as timer
	rep #$20                // 16-bit A
	lda.w #$0004            // Slow bit
	rtl
}

// set Shell and start timer

scope set_shell_ex: {
	sep #$20				// 8-bit A
	lda.b #shell_timer		// Shell timer value
	sta $3E88,y				// set as timer
	rtl
}

// set Safe and start timer

scope set_safe_ex: {
	sep #$20				// 8-bit A
	lda.b #safe_timer		// Safe timer value
	sta $3E89,y				// set as timer
	rtl
}

// relocated and expanded $C25B06 routine
// check if timers have expired

scope check_timers_ex: {
    lda.w $3F0C,x       // time until Reflect expires
    beq check_freeze    // branch if Reflect timer is inactive
    dec.w $3F0C,x       // decrement Reflect timer
    bne check_freeze    // branch if Reflect timer has not reached 0
    lda.b #$02          // Reflect bit
    tsb.b $B8           // set Reflect timer as expired
check_freeze:
    lda.w $3F0D,x       // time until Freeze expires
    beq check_sleep     // branch if Freeze timer is inactive
    dec.w $3F0D,x       // decrement Freeze timer
    bne check_sleep     // branch if Freeze timer has not reached 0
    lda.b #$04          // Freeze bit
    tsb.b $B8           // set Freeze timer as expired
check_sleep:
    lda.w $3CF9,x       // time until Sleep expires
    beq check_haste     // branch if Sleep timer is inactive
    dec.w $3CF9,x       // decrement Sleep timer
    bne check_haste     // branch if Sleep timer has not reached 0
    lda.b #$08          // Sleep bit
    tsb.b $B8           // set Sleep timer as expired
check_haste:
    lda.w $3E39,x       // time until Haste expires
    beq check_slow      // branch if Haste timer is inactive
    dec.w $3E39,x       // decrement Haste timer
    bne check_slow      // branch if Haste timer has not reached 0
    lda.b #$10          // Haste bit
    tsb.b $B8           // set Haste timer as expired
check_slow:
    lda.w $3E38,x       // time until Slow expires
    beq check_shell     // branch if Slow timer is inactive
    dec.w $3E38,x       // decrement Slow timer
    bne check_shell     // branch if Slow timer has not reached 0
    lda.b #$20          // Slow bit
    tsb.b $B8           // set Slow timer as expired
check_shell:
    lda.w $3E88,x       // time until Shell expires
    beq check_safe      // branch if Shell timer is inactive
    dec.w $3E88,x       // decrement Shell timer
    bne check_safe      // branch if Shell timer has not reached 0
    lda.b #$40          // Shell bit
    tsb.b $B8           // set Shell timer as expired
check_safe:
    lda.w $3E89,x       // time until Safe expires
    beq exit      		// branch if Safe timer is inactive
    dec.w $3E89,x       // decrement Safe timer
    bne exit      		// branch if Safe timer has not reached 0
    lda.b #$80          // Safe bit
    tsb.b $B8           // set Safe timer as expired
exit:
    lda.b $B8           // load current entity's expired status byte
    rtl
}

// expanded $C2518A routine
// remove status when timers have expired

scope remove_status_ex: {
    lsr.b $B8           // check if Sleep has expired
    bcc check_haste     // branch if not
    lda.b #$80          // Sleep bit
    and.w $3EE5,x
    beq check_haste     // branch if not sleeping
    tsb.w $11AB         // attack clears sleep
    lda.b #$02          // magic
    sta.b $B5           // set command to magic
    lda.b #$78          // Tapir spell ID
    sta.b $B6           // set spell to Tapir
check_haste:
    lsr.b $B8           // check if Haste has expired
    bcc check_slow      // branch if not
    lda.b #$08          // Haste bit
    tsb.w $11AC         // attack clears Haste
check_slow:
    lsr.b $B8           // check if Slow has expired
    bcc check_shell     // branch if not
    lda.b #$04          // Slow bit
    tsb.w $11AC         // attack clears Slow
check_shell:
    lsr.b $B8           // check if Shell has expired
    bcc check_safe      // branch if not
    lda.b #$20          // Shell bit
    tsb.w $11AC         // attack clears Shell
check_safe:
    lsr.b $B8           // check if Safe has expired
    bcc exit      		// branch if not
    lda.b #$40          // Safe bit
    tsb.w $11AC         // attack clears Safe
    exit:
    rtl
}